#! /usr/bin/env Rscript


#------------------------------------
# define functions to simulate data
#------------------------------------

#' makeData
#' 
#' Make the data used in the simulation study
#' 
#' @param n Sample size
#' @param beta Parameter dictating how informative the continuous covariate is.
#' @param tau The final timepoint
#' @param ... Other args, not currently used
#' 

makeData <- function(n, betaT = 2, betaC = 2, tau=6, ...){
  W1 <- runif(n, -2, 2)
  W2 <- rbinom(n, 1, 0.5)
  Z <- rbinom(n, 1, 0.5)

  # simulate data
  C <- rgeom(n,plogis(-3 + betaC*W1 - 2*betaC*W1*W2 + Z)) + 1
  T <- rgeom(n,plogis(-2 + betaT*W1 - 2*betaT*W1*W2 + Z)) + 1
  J <- rbinom(n,1,1/2) + 1

  # if everyone is censored at study end, fix up those times
  if(!is.na(tau)){
    C[C>tau] <- tau
  }

  # observed data time and event indicator
  ftime <- pmin(C,T)
  ftype <- 0*I(C<T) + 1*I(T <= C & J==1) + 2*I(T <= C & J==2) 

  # return a data frame
  rslt <- data.frame(id=1:n,
                   Z=Z,
                   W1=W1,
                   W2=W2,
                   ftime=ftime,
                   ftype=ftype)
  return(rslt)
}

#' getTruth
#' 
#' Get the true cumulative incidence in the vaccinated arm for
#' type 1 endpoints. 
#' @param n Sample size (should be large)
#' @param beta Interaction effect (should match makeData call)
#' @param t0 Time to compute incidence

getTruth <- function(n = 1e6, betaT = 2, t0 = 6){
  W1 <- runif(n, -2, 2)
  W2 <- rbinom(n, 1, 0.5)
  Z <- rbinom(n, 1, 0.5)

  # simulate data
  T <- rgeom(n,plogis(-2 + betaT*W1 - 2*betaT*W1*W2 + Z)) + 1
  J <- rbinom(n,1,1/2) + 1

  # true value
  Q0 <- sum(T <= t0 & Z ==1 & J==1)/sum(Z==1)
}

#' makeBiasData
#' 
#' Make the data used in the simulation study
#' 
#' @param n Sample size
#' @param beta Parameter dictating how informative the continuous covariate is.
#' @param tau The final timepoint
#' @param ... Other args, not currently used
#' 

makeBiasData <- function(n, betaT = 2, betaC = 2, tau=6, ...){
  W1 <- runif(n, -2, 2)
  W2 <- rbinom(n, 1, 0.5)
  U <- rnorm(n, W1, 1)
  Z <- rbinom(n, 1, 0.5)

  # simulate data
  C <- rgeom(n,plogis(-3 + betaC*(W1 - U) - 2*betaC*(W1 - U)*W2 + Z)) + 1
  T <- rgeom(n,plogis(-2 + betaT*(W1 - U) - 2*betaT*(W1 - U)*W2 + Z)) + 1
  J <- rbinom(n,1,1/2) + 1

  # if everyone is censored at study end, fix up those times
  if(!is.na(tau)){
    C[C>tau] <- tau
  }

  # observed data time and event indicator
  ftime <- pmin(C,T)
  ftype <- 0*I(C<T) + 1*I(T <= C & J==1) + 2*I(T <= C & J==2) 

  # return a data frame
  rslt <- data.frame(id=1:n,
                   Z=Z,
                   W1=W1,
                   W2=W2,
                   ftime=ftime,
                   ftype=ftype)
  return(rslt)
}

#' getBiasTruth
#' 
#' Get the true cumulative incidence in the vaccinated arm for
#' type 1 endpoints. 
#' @param n Sample size (should be large)
#' @param beta Interaction effect (should match makeData call)
#' @param t0 Time to compute incidence

getBiasTruth <- function(n = 1e6, betaT = 2, t0 = 6){
  W1 <- runif(n, -2, 2)
  W2 <- rbinom(n, 1, 0.5)
  U <- rnorm(n, W1, 1)
  Z <- rbinom(n, 1, 0.5)

  # simulate data
  T <- rgeom(n,plogis(-2 + betaT*(W1 - U) - 2*betaT*(W1 - U)*W2 + Z)) + 1
  J <- rbinom(n,1,1/2) + 1

  # true value
  Q0 <- sum(T <= t0 & Z ==1 & J==1)/sum(Z==1)
}


#--------------------------------------
# run simulation for first 4 settings
#--------------------------------------
ns <- c(500,5000) # sample sizes
bigB <- 1000 # number of replications
betaT <- c(0,2) # covariate relation to failure time
betaC <- c(0,2) # covariate relation to censoring
# make a grid of simulation parameters 
parm <- expand.grid(seed=1:bigB,
                    n=ns,
                    betaT = betaT,
                    betaC = betaC)

# load packages
library(nnls)
library(SuperLearner)
library(survtmle)
library(plotrix)
library(TeachingDemos)
library(plotmo)
library(earth)
library(survtmle)

# make a directory to save result files
saveDir <- "~/tmp"

# define a function that computes all estimators, formats 
# output and saves results to saveDir
do.one <- function(dat, parm, bias = FALSE){
  # set seed
  set.seed(parm$seed[i])

  # define super learner library      
  SL.lib <- c("SL.glm","SL.step.interaction","SL.mean","SL.earth")
    
  # fit the gcomp estimator
  gcomp <- survtmle(ftime = dat$ftime,
                     ftype = dat$ftype,
                     adjustVars = dat[,c("W1","W2")],
                     ftypeOfInterest = 1,
                     trtOfInterest = 1,
                     trt = dat$Z,
                     glm.trt = "1",
                     glm.ftime = "W1 + W2 + trt",
                     glm.ctime = "trt",
                     Gcomp = TRUE,
                     method = "mean",
                     t0=6)
  # compute ci for gcomp estimator 
  ci.gcomp <- rep(gcomp$est,2) + c(-1.96, 1.96)*rep(sqrt(gcomp$var),2)

  # compute the iterated mean tmle
  tmle1 <- survtmle(ftime = dat$ftime,
                     ftype = dat$ftype,
                     ftypeOfInterest = 1,
                     trtOfInterest = 1,
                     adjustVars = dat[,c("W1","W2")],
                     trt = dat$Z,
                     glm.trt = "1",
                     SL.ftime = SL.lib,
                     SL.ctime = SL.lib,
                     method = "mean",
                     t0=6,gtol=0.01)

  # compute ci for iterated mean tmle
  ci.tmle1 <- rep(tmle1$est,2) + c(-1.96, 1.96)*rep(sqrt(tmle1$var),2)

  # compte the hazard tmle 
  tmle2 <- survtmle(ftime = dat$ftime,
                 ftype = dat$ftype,
                 ftypeOfInterest = 1,
                 trtOfInterest = c(0,1),
                 adjustVars = dat[,c("W1","W2")],
                 trt = dat$Z,
                 glm.trt = "1",
                 SL.ftime = SL.lib,
                 SL.ctime = SL.lib,
                 method = "hazard", tol = 1/sqrt(length(dat$ftime)),
                 t0=6,verbose = FALSE,gtol=0.01)
  # compute the ci for hazard tmle
  ci.tmle2 <- rep(tmle2$est[2],2) + c(-1.96, 1.96)*rep(sqrt(tmle2$var[2,2]),2)

  # compute the AJ estimator, which is a TMLE that only adjusts for treatment
  uadj <- survtmle(ftime = dat$ftime,
                   ftype = dat$ftype,
                   ftypeOfInterest = 1,
                   trtOfInterest = 1,
                   adjustVars = dat[,c("W1","W2")],
                   trt = dat$Z,
                   glm.trt = "1",
                   glm.ftime = "trt",
                   glm.ctime = "trt",
                   method = "mean",
                   t0=6)
  # compute confidence interval for AJ
  ci.uadj <- rep(uadj$est,2) + c(-1.96, 1.96)*rep(sqrt(uadj$var),2)


  # find true parameter value
  set.seed(1248612)
  if(!bias){
    truth <- getTruth(betaT = parm$betaT[i])
  }else{
    truth <- getBiasTruth(betaT = parm$betaT[i])
  }
  # output formatted as:
  # truth, estimators, coverage of estimated CI
  out <- c(parm$seed[i],parm$n[i],truth, 
           parm$betaT[i],parm$betaC[i],
           uadj$est, ci.uadj, gcomp$est, ci.gcomp, tmle1$est, ci.tmle1, tmle2$est[2], ci.tmle2)
  fileName <- "out"
  if(bias){
    fileName <- "outBias"
  }
  save(out, file = paste0(saveDir,"/",fileName,"_n=",parm$n[i],"_seed=",parm$seed[i],
                         "_betaT=",parm$betaT[i],"_betaC=",parm$betaC[i],".RData"))
}

# this for loop runs the simulation for the first four settings
for(i in 1:nrow(parm)){
  # make data set
  set.seed(parm$seed[i])
  dat <- makeData(n=parm$n[i], betaT = parm$betaT[i], betaC = parm$betaC[i])
  # run estimators and save results
  do.one(dat, parm = parm, bias = FALSE)
}

#-------------------------------------------------------
# now re-run in the setting with unmeasured confounder
#-------------------------------------------------------
ns <- c(500, 5000)
bigB <- 1000
betaT <- c(2)
betaC <- c(2)

bias_parm <- expand.grid(seed=1:bigB,
                         n=ns,
                         betaT = betaT,
                         betaC = betaC)

for(i in 1:nrow(bias_parm)){
  # make data set
  set.seed(bias_parm$seed[i])
  dat <- makeBiasData(n=bias_parm$n[i], betaT = bias_parm$betaT[i], 
                      betaC = bias_parm$betaC[i])
  # run estimators and save results
  do.one(dat, parm = bias_parm, bias = TRUE)
}

#-------------------------------
# merge results
#-------------------------------
rslt <- NULL
for(i in 1:nrow(parm)){
    tmp <- tryCatch({
        load(paste0(saveDir,"/out_n=",parm$n[i],"_seed=",parm$seed[i],
                       "_betaT=",parm$betaT[i],"_betaC=",parm$betaC[i],".RData"))
        out
    }, error=function(e){
      c(parm$seed[i],parm$n[i], NA,
         parm$betaT[i],parm$betaC[i],rep(NA,12))
    })
    rslt <- rbind(rslt, tmp)
}
# format
out <- data.frame(rslt)
colnames(out) <- c(
    "seed","n","truth","betaT","betaC",
    c(t(outer(c("uadj","gcomp","tmle1","tmle2"),c("Est","CiL","CiU"),FUN=paste0)))
)
row.names(out) <- NULL
    
save(out, file=paste0(saveDir,"/allOut.RData"))

# merge results from sims with unmeasured confounder
rslt <- NULL
for(i in 1:nrow(bias_parm)){
    tmp <- tryCatch({
        load(paste0(saveDir,"/outBias_n=",bias_parm$n[i],"_seed=",bias_parm$seed[i],
                       "_betaT=",bias_parm$betaT[i],"_betaC=",bias_parm$betaC[i],".RData"))
        out
    }, error=function(e){
      c(bias_parm$seed[i],bias_parm$n[i], NA,
         bias_parm$betaT[i],bias_parm$betaC[i],rep(NA,12))
    })
    rslt <- rbind(rslt, tmp)
}
# format
out_bias <- data.frame(rslt)
colnames(out_bias) <- c(
    "seed","n","truth","betaT","betaC",
    c(t(outer(c("uadj","gcomp","tmle1","tmle2"),c("Est","CiL","CiU"),FUN=paste0)))
)
row.names(out_bias) <- NULL
    
save(out_bias, file=paste0(saveDir,"/allBiasOut.RData"))


#-------------------------------------------------------
# compute summary statistics on bias, cv, and coverage
#-------------------------------------------------------
load(paste0(saveDir,"/allOut.RData"))
colnames(out)[1:5] <- c("seed","n","truth","betaT","betaC")
out <- data.frame(out,
     uadjBias = out$uadjEst - out$truth,
     gcompBias = out$gcompEst - out$truth,
     tmle1Bias = out$tmle1Est - out$truth,
     tmle2Bias = out$tmle2Est - out$truth
)
# true values of cumulative incidence (for table)
truth_1 <- truth_2 <- out$truth[1]
truth_3 <- truth_4 <- out$truth[length(out$truth)]

## first compute summary statistics for n = 500 case
out500 <- out[out$n==500,]
# retrieve bias 
b500 <- by(out500,interaction(out500$betaT,out500$betaC), function(x){
  colMeans(x[,grep("Bias",colnames(out500))]/x$truth)
})
# retrieve coefficient of variation
cv500 <- by(out500,interaction(out500$betaT,out500$betaC), function(x){
  sd_est <- apply(x[,grep("Est",colnames(out500))],2,sd)
  mean_est <- apply(x[,grep("Est",colnames(out500))],2,mean)
  sd_est/mean_est
})
# compute coverage
c500 <- by(out500,interaction(out500$betaT,out500$betaC), function(x){
  tmp <- NULL
  for(est in c("uadj","gcomp","tmle1","tmle2")){
    tmp <- c(tmp, 
             mean(x[,paste0(est,"CiL")] < x$truth &
                  x[,paste0(est,"CiU")] > x$truth))
  }
  tmp
})

## now compute summary statistics for n = 5000 case
out5000 <- out[out$n==5000,]
# bias
b5000 <- by(out5000,interaction(out5000$betaT,out5000$betaC), function(x){
  colMeans(x[,grep("Bias",colnames(out5000))]/x$truth)
})
# coefficient of variation
cv5000 <- by(out5000,interaction(out5000$betaT,out5000$betaC), function(x){
  sd_est <- apply(x[,grep("Est",colnames(out5000))],2,sd)
  mean_est <- apply(x[,grep("Est",colnames(out5000))],2,mean)
  sd_est/mean_est
})
# coverage
c5000 <- by(out5000,interaction(out5000$betaT,out5000$betaC), function(x){
  tmp <- NULL
  for(est in c("uadj","gcomp","tmle1","tmle2")){
    tmp <- c(tmp, 
             mean(x[,paste0(est,"CiL")] < x$truth &
                  x[,paste0(est,"CiU")] > x$truth))
  }
  tmp
})

## now the same operations for the case with unmeasured confounders
colnames(out_bias)[1:5] <- c("seed","n","truth","betaT","betaC")
out_bias <- data.frame(out_bias,
     uadjBias = out_bias$uadjEst - out_bias$truth,
     gcompBias = out_bias$gcompEst - out_bias$truth,
     tmle1Bias = out_bias$tmle1Est - out_bias$truth,
     tmle2Bias = out_bias$tmle2Est - out_bias$truth
)
truth_5 <- out_bias$truth[1]

out500 <- out_bias[out_bias$n==500,]
# bias
b500_bias <- by(out500,interaction(out500$betaT,out500$betaC), function(x){
  colMeans(x[,grep("Bias",colnames(out500))]/x$truth)
})
# coefficient of variation
cv500_bias <- by(out500,interaction(out500$betaT,out500$betaC), function(x){
  sd_est <- apply(x[,grep("Est",colnames(out500))],2,sd)
  mean_est <- apply(x[,grep("Est",colnames(out500))],2,mean)
  sd_est/mean_est
})
# coverage
c500_bias <- by(out500,interaction(out500$betaT,out500$betaC), function(x){
  tmp <- NULL
  for(est in c("uadj","gcomp","tmle1","tmle2")){
    tmp <- c(tmp, 
             mean(x[,paste0(est,"CiL")] < x$truth &
                  x[,paste0(est,"CiU")] > x$truth))
  }
  tmp
})

out5000 <- out_bias[out_bias$n==5000,]
# bias
b5000_bias <- by(out5000,interaction(out5000$betaT,out5000$betaC), function(x){
  colMeans(x[,grep("Bias",colnames(out5000))]/x$truth)
})
# coefficient of variation
cv5000_bias <- by(out5000,interaction(out5000$betaT,out5000$betaC), function(x){
  sd_est <- apply(x[,grep("Est",colnames(out5000))],2,sd)
  mean_est <- apply(x[,grep("Est",colnames(out5000))],2,mean)
  sd_est/mean_est
})
# coverage
c5000_bias <- by(out5000,interaction(out5000$betaT,out5000$betaC), function(x){
  tmp <- NULL
  for(est in c("uadj","gcomp","tmle1","tmle2")){
    tmp <- c(tmp, 
             mean(x[,paste0(est,"CiL")] < x$truth &
                  x[,paste0(est,"CiU")] > x$truth))
  }
  tmp
})

#----------------------------------
# make latex table using xtable
#----------------------------------
require(xtable)

# check mark
cmark <- "\\checkmark"
xmark <- "\\xmark"
# bias multiplier
bmult <- 1000
vmult <- 10
mmult <- 10000
cmult <- 100

# setup dataframe
df <- data.frame(c(paste("$W \\not\\to T$, $W \\not\\to C$"), 
                   paste0("$F_{1,0}^1(6) = ", round(truth_1,2),"$"),"",
                 paste("$W \\not\\to T$, $W \\to C$"),
                   paste0("$F_{1,0}^1(6) = ", round(truth_2,2),"$"),"",
                 paste("$W \\to T$, $W \\not\\to C$"),
                   paste0("$F_{1,0}^1(6) = ", round(truth_3,2),"$"),"",
                 paste("$W \\to T$, $W \\to C$"),
                   paste0("$F_{1,0}^1(6) = ", round(truth_4,2),"$"),"",
                 paste("$(W,U) \\to T$, $(W,U) \\to C$"),
                   paste0("$F_{1,0}^1(6) = ", round(truth_5,2),"$"),""),
                 rep(c("Consistent","Efficient",""),5),
                 c(rep(cmark,7),rep(xmark,8)),
                 c(rep(cmark,7),rep(xmark,8)),
                 c(rep(cmark,12),rep(xmark,3)),
                 rep(c("Relative Bias (\\%)","Coef. Var. ($\\times$10)","Coverage (\\%)"),5),
                 rbind(b500[[1]][1:3]*100,cv500[[1]][1:3]*vmult,c500[[1]][1:3]*cmult,
                       b500[[3]][1:3]*100,cv500[[3]][1:3]*vmult,c500[[3]][1:3]*cmult,
                       b500[[2]][1:3]*100,cv500[[2]][1:3]*vmult,c500[[2]][1:3]*cmult,
                       b500[[4]][1:3]*100,cv500[[4]][1:3]*vmult,c500[[4]][1:3]*cmult,
                       b500_bias[[1]][1:3]*100,cv500_bias[[1]][1:3],c500_bias[[1]][1:3]*cmult),
                 rbind(b5000[[1]][1:3]*100,cv5000[[1]][1:3]*vmult,c5000[[1]][1:3]*cmult,
                       b5000[[3]][1:3]*100,cv5000[[3]][1:3]*vmult,c5000[[3]][1:3]*cmult,
                       b5000[[2]][1:3]*100,cv5000[[2]][1:3]*vmult,c5000[[2]][1:3]*cmult,
                       b5000[[4]][1:3]*100,cv5000[[4]][1:3]*vmult,c5000[[4]][1:3]*cmult,
                       b5000_bias[[1]][1:3]*100,cv5000_bias[[1]][1:3]*vmult,c5000_bias[[1]][1:3]*cmult)
                 )
for(i in 1:5){
  df[[i]] <- as.character(df[[i]])
}
rle.lengths <- rle(df[[1]])$lengths

first <- !duplicated(df[[1]])
df[[1]][!first] <- ""

# define appearance of \multirow
for(i in 1:5){
df[[i]][seq(1,length(df[[i]]),2)] <-
   paste0("\\multirow{", 2, "}{*}{", df[[i]][seq(1,length(df[[i]]),2)], "}")
df[[i]][seq(2,length(df[[i]]),2)] <-
   paste0("\\multirow{", 2, "}{*}{", df[[i]][seq(2,length(df[[i]]),2)], "}")
df[[i]][seq(3,length(df[[i]]),3)] <- ""
}


# set up xtable output
print(xtable(df, digits = matrix(rep(c(rep(0,7),rep(2,6),
                                   rep(0,7),rep(2,6),
                                   rep(0,7),rep(1,6)),5),nrow = 15, byrow=TRUE), # first zero "represents" row numbers which we skip later
             align = "ccccccccccccc",  # align and put a vertical line (first "l" again represents column of row numbers)
             label = "testTable"),
      size = "footnotesize", #Change size; useful for bigger tables "normalsize" "footnotesize"
      include.rownames = FALSE, #Don't print rownames
      include.colnames = FALSE, #We create them ourselves
      caption.placement = "top", #"top", NULL
      hline.after=seq(3,12,3), 
      floating=TRUE, # whether \begin{Table} should be created (TRUE) or not (FALSE)
      sanitize.text.function = force, # Important to treat content of first column as latex function
      add.to.row = list(pos = list(-1,
                                   nrow(df)),
                        command = c(paste("\\toprule \n"," & & \\multicolumn{3}{c}{Properties} & & \\multicolumn{3}{c}{$n=500$} & \\multicolumn{3}{c}{$n=5000$} \\\\\n", 
      "\\cmidrule{3-5}\\cmidrule(lr){7-9}\\cmidrule(lr){10-12}\n",
        "& & \\hphantom{A}AJ\\hphantom{A} & GCOMP & TMLE & & \\hphantom{A}AJ\\hphantom{A} & GCOMP & TMLE & \\hphantom{A}AJ\\hphantom{A} & GCOMP & TMLE \\\\\n",
        "\\cmidrule{3-5}\\cmidrule(lr){7-9}\\cmidrule(lr){10-12}"),
                                    paste("\\bottomrule \n")
                                    )
                        )
      )